/*
 * mesh.c
 */

#include "multilevel.h"
#include "mesh.h"
#include <math.h>


Vertex **vertices;
Triangle *triangles;
int vertices_length, vertices_alloc, triangles_length, triangles_alloc;

static Vertex*
alloc_vertex(float v[],
             float n[]) {
  if (vertices_alloc == 0) {
    vertices_alloc = 64;
    vertices = MALLOC(sizeof(Vertex*) * vertices_alloc);
  }
  if (vertices_length == vertices_alloc) {
    vertices_alloc *= 2;
    vertices = REALLOC(vertices, sizeof(Vertex*) * vertices_alloc);
  }
  vertices[vertices_length] = MALLOC(sizeof(Vertex));
  return vertices[vertices_length++];
}


#define VERTEX_TABLE_LENGTH 1024
Vertex* vertex_table[VERTEX_TABLE_LENGTH];

static int
hash_tag(int tag)
{
  return (tag ^ (tag >> 8) ^ (tag >> 16) ^ (tag >> 24)) % VERTEX_TABLE_LENGTH;
}

static int
put_vertex(int tag, Vertex *v)
{
  int hash = hash_tag(tag);
  v->next = vertex_table[hash];
  vertex_table[hash] = v;
  return tag;
}

Vertex*
add_vertex(int tag,
           float v[],
           float n[])
{
  Vertex *vert = alloc_vertex(v, n);

  vert->x = v[0];
  vert->y = v[1];
  vert->z = v[2];
  vert->nx = n[0];
  vert->ny = n[1];
  vert->nz = n[2];
  vert->cr = 0.8;
  vert->cg = 0.95;
  vert->cb = 0.95;

  vert->tag = tag;
  vert->adj = 0;
  vert->next = 0;

  vert->tris_length = 0;
  vert->tris_alloc = 6;
  vert->tris = MALLOC(vert->tris_alloc * sizeof(int));

  put_vertex(tag, vert);

  return vert;
}


Vertex*
get_vertex(int tag)
{
  Vertex *v = vertex_table[hash_tag(tag)];
  if (v == 0) return 0;
  while (v->tag != tag) {
    v = v->next;
    if (v == 0) return 0;
  }
  return v;
}

void
clear_vertices()
{
  int i;
  for (i = 0; i < vertices_length; i++) {
    vertices[i]->cr = 0.8;
    vertices[i]->cg = 0.95;
    vertices[i]->cb = 0.95;
    vertices[i]->adj = 0;
    vertices[i]->tris_length = 0;
  }
}


int
add_triangle_to_vertex(Vertex *v,
                       int t)
{
  if (v->tris_length == v->tris_alloc) {
    v->tris_alloc *= 2;
    v->tris = REALLOC(v->tris, sizeof(int) * v->tris_alloc);
  }
  v->tris[v->tris_length++] = t;
  return 0;
}

int
add_triangle(Vertex* t[]) {
  int k;

  if (triangles_alloc == 0) {
    triangles_alloc = 64;
    triangles = MALLOC(sizeof(Triangle) * triangles_alloc);
  }
  if (triangles_length == triangles_alloc) {
    triangles_alloc *= 2;
    triangles = REALLOC(triangles, sizeof(Triangle) * triangles_alloc);
  }
  triangles[triangles_length].v[0] = t[0];
  triangles[triangles_length].v[1] = t[1];
  triangles[triangles_length].v[2] = t[2];
  triangles[triangles_length].submeshes = 0x00000001;

  for (k = 0; k < 3; k++)
    add_triangle_to_vertex(t[k], triangles_length);

  return triangles_length++;
}

void
clear_triangles()
{
  triangles_length = 0;
}



int
triangle_normal(Vertex *t[3], float n[3])
{
  float v1[3], v2[3], v3[3];
  float ax, ay, az, bx, by, bz, r;

  v1[0] = t[0]->x;
  v1[1] = t[0]->y;
  v1[2] = t[0]->z;
  v2[0] = t[1]->x;
  v2[1] = t[1]->y;
  v2[2] = t[1]->z;
  v3[0] = t[2]->x;
  v3[1] = t[2]->y;
  v3[2] = t[2]->z;

  ax = v2[0] - v1[0];
  ay = v2[1] - v1[1];
  az = v2[2] - v1[2];
  bx = v3[0] - v1[0];
  by = v3[1] - v1[1];
  bz = v3[2] - v1[2];
  n[0] = - ay * bz + az * by;
  n[1] = - az * bx + ax * bz;
  n[2] = - ax * by + ay * bx;
  r = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
  if (r < 0.00001) {
    n[0] = n[1] = n[2] = 0.0;
    return 1;
  }
  n[0] /= r;
  n[1] /= r;
  n[2] /= r;
  return 0;
}


int
recompute_vertex_normals()
{
  int i, j;
  float n[3];
  for (i = 0; i < vertices_length; i++) {
    vertices[i]->nx = vertices[i]->ny = vertices[i]->nz = 0.0;
    vertices[i]->count = 0;
  }
  for (i = 0; i < triangles_length; i++) {
    triangle_normal(triangles[i].v, n);
    for (j = 0; j < 3; j++) {
#if 0
      n[0] = triangles[i].v[j]->x - triangles[i].v[(j+1)%3]->x;
      n[1] = triangles[i].v[j]->y - triangles[i].v[(j+1)%3]->y;
      n[2] = triangles[i].v[j]->z - triangles[i].v[(j+1)%3]->z;
#endif
      triangles[i].v[j]->nx += n[0];
      triangles[i].v[j]->ny += n[1];
      triangles[i].v[j]->nz += n[2];
      triangles[i].v[j]->count++;
    }
  }
  for (i = 0; i < vertices_length; i++) {
    vertices[i]->nx /= vertices[i]->count;
    vertices[i]->ny /= vertices[i]->count;
    vertices[i]->nz /= vertices[i]->count;
  }
  return 0;
}

float
triangle_area(Vertex *t[3])
{
  float dx, dy, dz, r1, r2, r3;

  dx = t[0]->x - t[1]->x;
  dy = t[0]->y - t[1]->y;
  dz = t[0]->z - t[1]->z;
  r1 = sqrt(dx*dx + dy*dy + dz*dz);

  dx = t[1]->x - t[2]->x;
  dy = t[1]->y - t[2]->y;
  dz = t[1]->z - t[2]->z;
  r2 = sqrt(dx*dx + dy*dy + dz*dz);

  dx = t[2]->x - t[0]->x;
  dy = t[2]->y - t[0]->y;
  dz = t[2]->z - t[0]->z;
  r3 = sqrt(dx*dx + dy*dy + dz*dz);

  return 0.25 * sqrt(2*r1*r2 + 2*r2*r3 + 2*r3*r1 - r1*r1 - r2*r2 - r3*r3);
}
