/*
 * voxels.c
 */

#include "multilevel.h"
#include "voxels.h"

#include <math.h>

uint8 voxels[DIMX+1][DIMY+1][DIMZ+1];


#define RAMP 4.0


static int
init_voxels_cube()
{
  int x, y, z;
  int dx, dy, dz, max;

  for (x = 0; x < DIMX+1; x++) {
    dx = x - DIMX/2;
    if (dx < 0) dx = -dx;

    for (y = 0; y < DIMY+1; y++) {
      dy = y - DIMY/2;
      if (dy < 0) dy = -dy;

      for (z = 0; z < DIMZ+1; z++) {
        dz = z - DIMZ/2;
        if (dz < 0) dz = -dz;

        max = dx;
        if (dy > max) max = dy;
        if (dz > max) max = dz;

        if        (max > DIMX/4 + detail*RAMP/2.0) {
          voxels[x][y][z] = 0;
        } else if (max < DIMX/4 - detail*RAMP/2.0) {
          voxels[x][y][z] = 255;
        } else {
          voxels[x][y][z]
            = (1.0 - (max - (DIMX/4 - detail*RAMP/2.0))/(detail*RAMP)) * 255.0;
        }
      }
    }
  }
  return 0;
}

static int
init_voxels_sphere()
{
  int x, y, z;
  float r;
  for (x = 0; x < DIMX+1; x++) {
    for (y = 0; y < DIMY+1; y++) {
      for (z = 0; z < DIMZ+1; z++) {
        r = sqrt((x-DIMX/2)*(x-DIMX/2) +
                 (y-DIMY/2)*(y-DIMY/2) +
                 (z-DIMZ/2)*(z-DIMZ/2));
        if        (r > DIMX/4 + detail*RAMP/2.0) {
          voxels[x][y][z] = 0;
        } else if (r < DIMX/4 - detail*RAMP/2.0) {
          voxels[x][y][z] = 255;
        } else {
          voxels[x][y][z]
            = (1.0 - (r - (DIMX/4 - detail*RAMP/2.0))/(detail*RAMP)) * 255.0;
        }
      }
    }
  }
  return 0;
}

static int
voxel_union_sphereflake(int depth, float c[], float n[])
{
  return 0;
}

static int
init_voxels_sphereflake(int depth)
{
  int x, y, z;
  float c[3], n[3];

  for (x = 0; x < DIMX+1; x++)
    for (y = 0; y < DIMY+1; y++)
      for (z = 0; z < DIMZ+1; z++)
        voxels[x][y][z] = 0;

  c[0] = c[1] = c[2] = 0.5;
  n[0] = 0.0;
  n[1] = 1.0;
  n[2] = 0.0;
  voxel_union_sphereflake(depth,   c, n);
  c[1] = 0.25;
  n[1] = -1.0;
  voxel_union_sphereflake(depth-1, c, n);
  return 0;
}

int
init_voxels()
{
  return init_voxels_cube();
}


int
voxel_eval(float v[3])
{
  int iv[3];
  float delta[3];
  int i, j, k;
  float value = 0.0;

  iv[0] = v[0] * (DIMX+1);
  iv[1] = v[1] * (DIMY+1);
  iv[2] = v[2] * (DIMZ+1);
  assert(iv[0] <= DIMX);
  assert(iv[1] <= DIMX);
  assert(iv[2] <= DIMX);

  delta[0] = v[0] * (DIMX+1) - iv[0];
  delta[1] = v[1] * (DIMY+1) - iv[1];
  delta[2] = v[2] * (DIMZ+1) - iv[2];

  for (i = 0; i < 2; i++)
    for (j = 0; j < 2; j++)
      for (k = 0; k < 2; k++)
        value +=
            (i ? delta[0] : (1.0-delta[0]))
          * (j ? delta[1] : (1.0-delta[1]))
          * (k ? delta[2] : (1.0-delta[2]))
          * voxels[iv[0]+i][iv[1]+j][iv[2]+k];

  return (int)value;
}


int
voxel_gradient(float  xx, float  yy, float  zz,
               float *nx, float *ny, float *nz)
{
  int x, y, z;
  int dfdx, dfdy, dfdz;

  x = xx * (float)DIMX;
  y = yy * (float)DIMY;
  z = zz * (float)DIMZ;

  if (x >= DIMX) {
    dfdx = 2*(voxels[x][y][z] - voxels[x-1][y][z]);
  } else if (x <= 0) {
    dfdx = 2*(voxels[x+1][y][z] - voxels[x][y][z]);
  } else {
    dfdx = voxels[x+1][y][z] - voxels[x-1][y][z];
  }
  if (y >= DIMY) {
    dfdy = 2*(voxels[x][y][z] - voxels[x][y-1][z]);
  } else if (y <= 0) {
    dfdy = 2*(voxels[x][y+1][z] - voxels[x][y][z]);
  } else {
    dfdy = voxels[x][y+1][z] - voxels[x][y-1][z];
  }
  if (z >= DIMZ) {
    dfdz = 2*(voxels[x][y][z] - voxels[x][y][z-1]);
  } else if (z <= 0) {
    dfdz = 2*(voxels[x][y][z+1] - voxels[x][y][z]);
  } else {
    dfdz = voxels[x][y][z+1] - voxels[x][y][z-1];
  }

  *nx = (float)dfdx / 2.0;
  *ny = (float)dfdy / 2.0;
  *nz = (float)dfdz / 2.0;

  return 0;
}

int
voxel_normal(float  xx, float  yy, float  zz,
             float *nx, float *ny, float *nz)
{
  float r;

  voxel_gradient(xx, yy, zz, nx, ny, nz);

  r = sqrt((*nx)*(*nx) + (*ny)*(*ny) + (*nz)*(*nz));
  if (r < 0.00001) {
    /* Don't divide by zero */
    *nx = *ny = *nz = 0.0;
    return 1;
  }
  *nx /= r;
  *ny /= r;
  *nz /= r;
  return 0;
}
